///////////////////////////////////////////////////////////////////////////////
// Librairie pour tracer des Cubes de type Rubik
// Fabrice Eudes
//
// version 2.0 - 15/03/2010
// - Initialisation avec la taille seule, ou détaillée.
// - Affichage de toutes les faces ou seulement des faces visibles.
// - Affichage remanié des faces des cubes élémentaires
// - Changement de l'origine pour les tracés : le cube est posé sur (O,X,Y,Z) alors
//   qu'avant il était centré en O.
// - Possibilité de choisir les couleurs et la taille des vignettes.
//
// version 1.0 - 09/02/2010
//
// A faire :
// - Placement du cube : triple pos
// - Rotation du cube : triple rot=(rotX,rotY,rotZ)
// (à vue de nez, il suffira de changer DessinerFaceCube1() et de modifications
// mineures ailleurs -passage des paramètres-).
///////////////////////////////////////////////////////////////////////////////

////
//// Chargement des modules utilisés
////
import three;
import roundedpath;

////
//// Constantes
////

// faces avant, droite, haut, arriere, gauche, bas.
pen[] CouleursFaces={blue, red, yellow, 0.5*green, orange, 0.95*white};
// taille des étiquettes, 0<T<1.
real T=0.9;
// Ne pas changer la ligne suivante !
triple[] NormalesFaces={X,Y,Z,-X,-Y,-Z};

////
//// Structures utilisateurs
////

//
// 1. un cube élémentaire
//
// triple avant -> normale orientant la face avant du cube.
// triple droite -> normale orientant la face droite du cube.
//
// Remarques :
// - valeurs à utiliser : X,-X,Y,-Y,Z,-Z
// - l'orientation de la face haut est le produit vectoriel direct cross(avant,droite)
//
struct Cube1 {
  triple avant;
  triple droite;

  // Initialisation par défaut
  void operator init() {
    this.avant=X;
    this.droite=Y;
  }
  // Initialisation alternative : on précise les orientations
  void operator init(triple avant, triple droite) {
    this.avant=avant;
    this.droite=droite;
  }
}

//
// 2. un Rubik's cube.
//
// int taille -> côté du cube en nombre de cubes élémentaires Cube1
// Cube1[] cubes -> le tableau des taille^3 Cube1 constituant le Rubik's cube
//
struct Cube {
  int taille;
  Cube1[] cubes;

  // Initialisation par défaut : on donne la taille du cube
  void operator init(int taille) {
    this.taille=taille;
    for (int i=0; i < this.taille^3; ++i) {
      this.cubes[i]=Cube1(X,Y);
    }
  }
  // Initialisation alternative : on donne aussi les orientations de chacun des Cube1
  void operator init(int taille, Cube1[] cubes) {
    this.taille=taille;
    this.cubes=cubes;
  }
}

////
//// Procédures de tracé
////

//// 1. Cube élémentaire
//
// Dessiner une face
//
// triple n -> vecteur normal à la face
// pen couleur -> couleur de la face
//
picture DessinerFaceCube1(triple n, pen couleur){
  path3 fond, dessus;
  triple normal;
  if (n == Z) {
    fond=shift(n)*unitsquare3;
    dessus=shift(1.01*n + (1-T)/2*(X+Y))*path3(roundedpath(unitsquare,1-T,S=T));
    }
  else if (n == -Z) {
    fond=unitsquare3;
    dessus=shift(0.01*n + (1-T)/2*(X+Y))*path3(roundedpath(unitsquare,1-T,S=T));
  }
  else if (n == X) {
    fond=shift(n)*rotate(90,cross(n,Z))*unitsquare3;
    dessus=shift(1.01*n + (1-T)/2*(Y+Z))*rotate(90,cross(n,Z))*path3(roundedpath(unitsquare,1-T,S=T));
  }
  else if (n == -X) {
    fond=rotate(-90,cross(n,Z))*unitsquare3;
    dessus=shift(0.01*n + (1-T)/2*(Y+Z))*rotate(-90,cross(n,Z))*path3(roundedpath(unitsquare,1-T,S=T));
  }
  else if (n == Y) {
    fond=shift(n)*rotate(90,cross(n,Z))*unitsquare3;
    dessus=shift(1.01*n + (1-T)/2*(X+Z))*rotate(90,cross(n,Z))*path3(roundedpath(unitsquare,1-T,S=T));
  }
  else {
    fond=rotate(-90,cross(n,Z))*unitsquare3;
    dessus=shift(0.01*n + (1-T)/2*(X+Z))*rotate(-90,cross(n,Z))*path3(roundedpath(unitsquare,1-T,S=T));
  }
  picture pic;
  draw(pic,surface(fond),black);
  draw(pic,surface(dessus),couleur);
  return(pic);
};

//
// Dessiner les six faces
//
// Cube1 c -> deux "triple" correspondants aux normales respectives des faces avant et droite.
// La normale à la face haut est déduite.
//
picture DessinerCube1(Cube1 c){
  picture pic;
  add(pic,DessinerFaceCube1(c.avant,CouleursFaces[0]));
  add(pic,DessinerFaceCube1(c.droite,CouleursFaces[1]));
  add(pic,DessinerFaceCube1(cross(c.avant,c.droite),CouleursFaces[2]));
  add(pic,DessinerFaceCube1(-c.avant,CouleursFaces[3]));
  add(pic,DessinerFaceCube1(-c.droite,CouleursFaces[4]));
  add(pic,DessinerFaceCube1(-cross(c.avant,c.droite),CouleursFaces[5]));
  return(pic);
}

//// 2. Rubik's cube
//
// Dessiner une face complète
//
// Cube c -> le Rubik's cube
// int face -> la face à dessiner
//             0 -> avant
//             1 -> droite
//             2 -> haut
//             3 -> arriere
//             4 -> gauche
//             5 -> bas
//
picture DessinerFaceCube(Cube c, int face){
  picture pic;
  Cube1 cc;
  int u,v,w;
  // Boucles sur les taille^2 cubes élémentaires de la face
  for (int i=0; i<c.taille; ++i) {
    for (int j=0; j<c.taille; ++j) {
      // cas face avant
      if (face == 0) {
        u=c.taille-1;
        v=i;
        w=j;
      }
      // cas face droite
      else if (face == 1) {
        u=i;
        v=c.taille-1;
        w=j;
      }
      // cas face haut
      else if (face == 2) {
        u=i;
        v=j;
        w=c.taille-1;
      }
      // cas face arriere
      else if (face == 3) {
        u=0;
        v=i;
        w=j;
      }
      // cas face gauche
      else if (face == 4) {
        u=i;
        v=0;
        w=j;
      }
      // cas face bas
      else {
        u=i;
        v=j;
        w=0;
      }
      // Choix du cube élémentaire
      cc=c.cubes[c.taille^2*u+c.taille*v+w];
      // Choix de la face du cube élémentaire
      if (NormalesFaces[face] == cc.avant) add(pic,shift(u*X+v*Y+w*Z)*DessinerFaceCube1(NormalesFaces[face],CouleursFaces[0]));
      else if (NormalesFaces[face] == cc.droite) add(pic,shift(u*X+v*Y+w*Z)*DessinerFaceCube1(NormalesFaces[face],CouleursFaces[1]));
      else if (NormalesFaces[face] == cross(cc.avant,cc.droite)) add(pic,shift(u*X+v*Y+w*Z)*DessinerFaceCube1(NormalesFaces[face],CouleursFaces[2]));
      else if (NormalesFaces[face] == -cc.avant) add(pic,shift(u*X+v*Y+w*Z)*DessinerFaceCube1(NormalesFaces[face],CouleursFaces[3]));
      else if (NormalesFaces[face] == -cc.droite) add(pic,shift(u*X+v*Y+w*Z)*DessinerFaceCube1(NormalesFaces[face],CouleursFaces[4]));
      else add(pic,shift(u*X+v*Y+w*Z)*DessinerFaceCube1(NormalesFaces[face],CouleursFaces[5]));
    }
  }
  return(pic);
}

//
// Dessiner les faces visibles d'un Rubik's cube (6*taille^2 faces).
//
picture DessinerFacesCube(Cube c) {
  picture pic;
  for (int f=0; f<6; ++f) {
    add(DessinerFaceCube(c,f));
  }
  return(pic);
}

//
// Dessiner tous les cubes élémentaires (6*taille^3 faces). ATTENTION : long !
//
picture DessinerCube(Cube c) {
  picture pic;
  for (int i=0; i<c.taille; ++i) {
    for (int j=0; j<c.taille; ++j) {
      for (int k=0; k<c.taille; ++k) {
        add(pic,shift(i*X+j*Y+k*Z)*DessinerCube1(c.cubes[c.taille^2*i+c.taille*j+k]));
      }
    }
  }
  return(pic);
}


////
//// Procédures de manipulation
////
//
// 1. cube élémentaire : rotation
//
// Cube1 c -> le cube
// triple axe -> l'axe de rotation (90°, sens direct)
//
Cube1 RotationCube1(Cube1 c, triple axe){
  if (cross(axe,c.avant) != O) c.avant=cross(axe,c.avant);
  if (cross(axe,c.droite) != O) c.droite=cross(axe,c.droite);
  return c;
}

//
// 2. Rubik's cube : rotation d'une tranche
//
// Cube c -> le Rubik's cube
// triple axe -> l'axe de rotation (90°, sens direct)
// int tranche -> l'ordre de la tranche, 0 <= tranche <= c.taille - 1;
//
Cube RotationCube(Cube c, triple axe, int tranche){
  // Rotation d'axe X
  if (axe == X) {
    // 1. Chaque cube de la tranche pivote d'un quart de tour
    for (int j=0; j<c.taille; ++j) {
      for (int k=0; k<c.taille; ++k) {
        RotationCube1(c.cubes[c.taille^2*tranche+c.taille*j+k],axe);
      }
    }
    // 2. Les cubes de la tranche changent de place :
    // on recopie les cubes élémentaires la constituant dans un tableau intermédiaire en ajustant l'ordre,
    // puis on recopie ce tableau intermédiaire dans la tranche d'origine
    Cube1[] cc;
    for (int j=0; j<c.taille; ++j) {
      for (int k=0; k<c.taille; ++k) {
        cc[c.taille^2*tranche+c.taille*j+k]=c.cubes[c.taille^2*tranche+c.taille*k+(c.taille-(j+1))];
      }
    }
    for (int j=0; j<c.taille; ++j) {
      for (int k=0; k<c.taille; ++k) {
        c.cubes[c.taille^2*tranche+c.taille*j+k]=cc[c.taille^2*tranche+c.taille*j+k];
      }
    }
    return(c);
  }
  // Rotation d'axe Y
  else if (axe == Y) {
    for (int i=0; i<c.taille; ++i) {
      for (int k=0; k<c.taille; ++k) {
        RotationCube1(c.cubes[c.taille^2*i+c.taille*tranche+k],axe);
      }
    }
    Cube1[] cc;
    for (int i=0; i<c.taille; ++i) {
      for (int k=0; k<c.taille; ++k) {
        cc[c.taille^2*i+c.taille*tranche+k]=c.cubes[c.taille^2*(c.taille-(k+1))+c.taille*tranche+i];
      }
    }
    for (int i=0; i<c.taille; ++i) {
      for (int k=0; k<c.taille; ++k) {
        c.cubes[c.taille^2*i+c.taille*tranche+k]=cc[c.taille^2*i+c.taille*tranche+k];
      }
    }
    return(c);
  }
  // Rotation d'axe Z
  else {
    for (int i=0; i<c.taille; ++i) {
      for (int j=0; j<c.taille; ++j) {
        RotationCube1(c.cubes[c.taille^2*i+c.taille*j+tranche],axe);
      }
    }
    Cube1[] cc;
    for (int i=0; i<c.taille; ++i) {
      for (int j=0; j<c.taille; ++j) {
        cc[c.taille^2*i+c.taille*j+tranche]=c.cubes[c.taille^2*j+c.taille*(c.taille-(i+1))+tranche];
      }
    }
    for (int i=0; i<c.taille; ++i) {
      for (int j=0; j<c.taille; ++j) {
        c.cubes[c.taille^2*i+c.taille*j+tranche]=cc[c.taille^2*i+c.taille*j+tranche];
      }
    }
    return(c);
  }
}
